import {file_url, system_file} from "../nodes.js";

import route_vue_set from "../route/vue/set.js";

import nodes from "../nodes.js";

import alone from "../alone.js";

import path from "path";

export default async function (dir = 'src', option = {}) {
    const self = {
        //开发目录
        dir: dir || 'src',
        //根目录
        pathname: (option.pathname || process.cwd()),
        //复制的目录
        copy_dir: (option.copy_dir || file_url(import.meta.url)),
        //当前目录
        dirname: file_url(import.meta.url),
        //配置
        option: Object.assign({
            //模式
            mode: "development",
            //打包还是服务
            command: "build",
            //开发目录
            alone: "alone",
            //临时信息
            cache: "file/cache.json",
            //配置文件名
            conf: "config/conf.js",
            //js文件名
            env: "config/env.js",
            //env.js开发key
            dev: "_dev_",
            //App.vue文件名
            app: "App.vue",
            //main.js文件名
            main: "main.js",
        }, option || {}),
        //判断是否/开头
        is_run(path) {
            return path.substring(0, 1) == '/';
        },
        vendor: {
            format: {
                img: ['jpeg', 'jpg', 'png', 'gif', 'svg', 'svgz', 'bmp', 'tiff', 'webp', 'heif', 'heic', 'ico'],
                font: ['ttf', 'otf', 'woff', 'woff2', 'eot'],
                audio: ['mp3', 'wav', 'ogg', 'aac', 'flac'],
                video: ['mp4', 'webm', 'ogv', 'avi', 'mov']
            },
            config: {
                //格式 包含有'svg', 'svgz',生成vendor.svg, svg或者file体积太大时可以选择开启
                svg: ['svg', 'svgz'],
                //路径包含有/vue-,生成vendor.vue, chunk体积太大时可以选择开启
                vue: function (file, format) {
                    if (file.includes('/vue-') && format.includes('js')) {
                        return true;
                    }
                },
                //路径包含有/alone/,生成vendor.alone
                alone: function (file, format) {
                    if (file.includes('/alone') && format.includes('js')) {
                        return true;
                    }
                },
                //路径包含有/naive-ui,生成vendor.naive, 推荐开启
                naive: function (file, format) {
                    if (file.includes('/naive-ui') && format.includes('js')) {
                        return true;
                    }
                },
                //格式 包含图片,字体,音频,视频打包成 vendor.file, 推荐开启
                file: function (file, format) {
                    if (this.vendor.format_arr.includes(format)) {
                        return true;
                    }
                },
                //路径包含有node_modules helper 打包成vendor.chunk, 推荐开启
                chunk(file, format) {
                    const chunk = ['node_modules', 'helpers', '-helper'];
                    if (chunk.some(val => file.includes(val))) {
                        return true;
                    }
                }
            }
        },
        env: {app_ver: String.fromCharCode(Math.floor(Math.random() * 26) + 97) + alone.rand_str()},
        //加载文件
        async import_file(file) {
            return await import(system_file(file));
        },
        //jsconfig.json配置
        js_config(object, path, pub, view, paths = {}) {
            paths["~/*"] = ["./*"];
            paths["&/*"] = ["./" + alone.trim(pub, '/') + "/*", "./" + path + "/" + alone.trim(pub, '/') + "/*"];
            paths["%/*"] = ["./" + path + "/" + view + "/*"];
            paths["@/*"] = ["./" + path + "/*"];
            paths["!/*"] = ["./" + alone.trim(this.option.alone, '/') + "/*"];
            object.compilerOptions = object.compilerOptions || {};
            object.compilerOptions.paths = paths;
            object.include = ["./" + path + "/**/*"];
            return object;
        },
        //启动
        async start() {
            this.update = false;
            //是否开发模式
            this.mode = this.option.mode == 'development';
            //当前目录node
            this.vite = nodes(this.dirname);
            //根目录node
            this.node = nodes(this.pathname);
            //复制目录node
            this.copy = nodes(this.copy_dir);
            //开发目录
            this.env.app_path = alone.trim(this.dir.replace(/\\/g, '/'), '/');
            //开发目录node
            this.app = nodes(this.node.get_path(this.env.app_path));
            //复制env
            this.option.env = alone.trim(this.option.env.replace(/\\/g, '/'), '/');
            if (!this.app.is_exist(this.option.env)) {
                this.app.cp_file(this.copy.get_path("config/env.js"), this.option.env, false);
            }
            //env内容
            this.conf_env = (await this.import_file(this.app.get_path(this.option.env))).default || {};
            //处理env
            Object.keys(this.conf_env).forEach((k) => {
                if (this.mode) {
                    const key = this.option.dev + k;
                    if (this.conf_env.hasOwnProperty(key)) {
                        this.env[k] = this.conf_env[key];
                    } else if (k.substring(0, this.option.dev.length) == this.option.dev) {
                        this.env[k.substring(this.option.dev.length)] = this.conf_env[k];
                    } else if (!this.env.hasOwnProperty(k)) {
                        this.env[k] = this.conf_env[k];
                    }
                } else if (k.substring(0, this.option.dev.length) != this.option.dev) {
                    this.env[k] = this.conf_env[k];
                }
            });
            //视图目录
            this.env.view = alone.trim(this.env.view, '/');
            //打包目录
            if (this.is_run(this.env.dist)) {
                this.env.dists = this.env.dist;
            } else {
                this.env.dist = alone.trim(this.env.dist, '/');
                this.env.dists = this.node.get_path(this.env.dist);
            }
            //静态目录
            this.env.public = alone.trim((this.env.public || 'public').replace(/\\/g, '/'), '/');
            //打包配置
            this.env.build = this.env.build || {};
            //临时数据
            this.cache = this.node.get_json(this.option.cache);
            if (!alone.isset(this.cache) || this.cache.path != this.env.app_path || this.cache.view != this.env.view) {
                this.update = true;
                this.vite.save_json(this.option.cache, {path: this.env.app_path, view: this.env.view});
            }
            //jsconfig.json
            this.is_jsconfig = this.node.is_exist('jsconfig.json');
            if (this.update || !this.is_jsconfig) {
                this.jsconfig_object = !this.is_jsconfig ? this.vite.get_json('file/jsconfig.json') : this.node.get_json('jsconfig.json');
                this.jsconfig_object = this.js_config(this.jsconfig_object, this.env.app_path, this.env.public, this.env.view);
                this.node.save_json('jsconfig.json', this.jsconfig_object);
            }
            //tsconfig.json
            this.is_tsconfig = this.node.is_exist('tsconfig.json');
            if (!this.is_tsconfig) {
                this.node.save_json('tsconfig.json', this.vite.get_json('file/tsconfig.json'));
            }
            //App.vue不存在时
            if (!this.app.is_exist(this.option.app)) {
                this.update = true;
                this.app.cp_file(this.copy.get_path("App.vue"), this.option.app, false);
                this.app.cp_file(this.copy.get_path("main.js"), this.option.main, false);
                this.app.cp_file(this.copy.get_path("config/conf.js"), this.option.conf, false);
                this.app.cp_dir_file(this.copy.get_path("view"), this.env.view, [], false);
                this.app.cp_dir_file(this.copy.get_path("mock"), 'mock', [], false);
                this.app.cp_dir_file(this.copy.get_path("style"), 'style', [], false);
            }
            //复制 layui
            if (!this.node.is_exist(this.env.public + "/layui/layui.js")) {
                this.node.cp_dir_file(this.vite.get_path("file/layui"), this.env.public + "/layui", [], false);
            }
            //项目public复制到public
            if (this.app.is_exist(this.env.public)) {
                this.node.cp_file(this.app.get_path(this.env.public), this.env.public, false);
            }
            //favicon.ico
            if (this.env.favicon && !this.is_run(this.env.favicon)) {
                this.env.favicon = (this.env.app_path + "/" + alone.trim(alone.trim(this.env.favicon, '.'), '/'));
                this.env.favicon = this.app.is_exist(this.env.favicon) ? this.env.favicon : "";
            }
            //conf内容
            this.config_conf = (this.app.is_exist(this.option.conf) ? ((await this.import_file(this.app.get_path(this.option.conf))).default) : {}) || {};
            //处理env
            this.del_env = ['host', 'port', 'build', 'zip', 'public', 'mock', 'favicon', 'view', 'app_path', 'dist'];
            this.config_env = {...this.env};
            this.del_env.forEach((k) => ((this.config_env.hasOwnProperty(k)) && delete this.config_env[k]));
            //文件格式 array
            this.vendor.format_arr = [].concat(...Object.values(this.vendor.format));
            //生成路由
            route_vue_set(this.app.get_path(this.env.view), {run: true});
            //打包配置
            this.vendor_config = {svg: false, vue: false, alone: true, naive: true, file: true, chunk: true};
            Object.keys(this.vendor.config).forEach((k) => {
                if (this.env.build[k] === true) {
                    this.vendor_config[k] = this.vendor.config[k];
                } else if (alone.isset(this.env.build[k])) {
                    this.vendor_config[k] = this.env.build[k];
                } else if (!this.env.build.hasOwnProperty(k) && this.vendor_config[k] === true) {
                    this.vendor_config[k] = this.vendor.config[k];
                } else {
                    delete this.vendor_config[k];
                }
            });
            if (alone.isset(this.env.build.callback) && alone.is_object(this.env.build.callback)) {
                Object.keys(this.env.build.callback).forEach((key) => {
                    if (alone.isset(this.env.build.callback[key])) {
                        this.vendor_config[key] = this.env.build.callback[key];
                    }
                });
            }
            return this;
        },
        publicDir() {
            return this.node.get_path(this.env.public);
        },
        define() {
            return {'process.env': this.config_env, 'process.config': this.config_conf};
        },
        resolve() {
            return {
                alias: {
                    '~': this.node.get_path('.'),
                    '&': this.node.get_path(this.env.public),
                    '@': this.app.get_path('.'),
                    '%': this.app.get_path(this.env.view),
                    '!': this.node.get_path(alone.trim(this.option.alone, '/'))
                }
            };
        },
        preview() {
            return {host: this.env.host || '127.0.0.1', port: this.env.port || '8088'};
        },
        server() {
            return Object.assign({
                proxy: {
                    '/api': {
                        target: this.config_env.api,
                        changeOrigin: true,
                        rewrite: (path) => path.replace(/^\/api/, '')
                    }
                }
            }, this.preview());
        },
        build() {
            return {
                outDir: this.env.dist,
                target: 'esnext',//'esnext', es2015 es2019
                minify: 'terser',//'terser',esbuild
                cssCodeSplit: true, // 如果设置为false，整个项目中的所有 CSS 将被提取到一个 CSS 文件中
                reportCompressedSize: false,//  关闭生成map文件 可以达到缩小打包体积
                chunkSizeWarningLimit: 1500,
                sourcemap: false, // 这个生产环境一定要关闭，不然打包的产物会很大
                terserOptions: {
                    format: {
                        beautify: false, // 不格式化输出
                        comments: false, // 不保留注释
                    },
                    mangle: {
                        toplevel: true, // 混淆顶级变量和函数
                    },
                    compress: {
                        properties: true,
                        dead_code: true,    // 删除未使用的代码
                        drop_console: true,//打包时删除console
                        drop_debugger: true,//打包时删除 debugger
                    },
                    sourceMap: true, // 生成源映射文件
                    ecma: 2015,      // 使用 ECMAScript 2015 的特性
                    enclose: false,  // 不包裹代码在闭包中
                    keep_classnames: false, // 不保留类名
                    keep_fnames: false, // 不保留函数名
                    ie8: false,       // 不兼容 IE8
                    module: false,    // 不支持 ES6 模块语法
                    nameCache: null,  // 不使用名称缓存
                    safari10: false,  // 不兼容 Safari 10
                    toplevel: false,  // 不混淆顶级作用域的变量
                },
                rollupOptions: {
                    treeshake: true, // 开启 Tree Shaking，消除未使用的代码，减小最终的包大小
                    output: {
                        hashCharacters: 'hex',
                        chunkFileNames: (obj) => {
                            const name = (obj.name).toString().toLowerCase()
                            return 'assets/js/' + name + '.[hash].js';
                        },
                        entryFileNames: (obj) => {
                            const name = (obj.name).toString().toLowerCase()
                            return 'assets/js/' + name + '.[hash].js';
                        },
                        assetFileNames: (obj) => {
                            const type = obj.type.toString().toLowerCase();
                            if (type === 'asset' && (obj.name || obj.names)) {
                                const file = (obj.name || (obj.names.slice(0, 1).join(''))).toString();
                                const files = file.replace(/\\/g, '/').toLowerCase();
                                const format = path.extname(files).slice(1);
                                const array = Object.keys(this.vendor.format);
                                for (const key of array) {
                                    const val = this.vendor.format[key];
                                    if (val && ((typeof val == 'function') ? val(files, format) : (Array.isArray(val) ? val.includes(format) : files.includes(val)))) {
                                        return 'assets/' + key + '/[name].[hash].[ext]';
                                    }
                                }
                            }
                            return 'assets/[ext]/[name].[hash].[ext]';
                        },
                        manualChunks: (file) => {
                            const files = file.replace(/\\/g, '/').toLowerCase();
                            const format = path.extname(files).slice(1);
                            const array = Object.keys(this.vendor_config);
                            for (const key of array) {
                                const val = this.vendor_config[key];
                                if (val && ((typeof val == 'function') ? val(files, format) : (Array.isArray(val) ? val.includes(format) : files.includes(val)))) {
                                    return "vendor." + key;
                                }
                            }
                        }
                    }
                }
            };
        }
    };
    return (await self.start());
}